# -*- coding: utf-8 -*-
#
#  Copyright 2020, 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from ...exceptions import EmptyMolecule


class EMOLRead:
    def __init__(self, ignore=False, log_buffer=None):
        self.__atoms = []
        self.__bonds = []
        self.__atom_map = {}
        self.__stereo = []
        self.__meta = {}
        self.__hydrogens = {}
        self.__star_points = []
        self.__ignore = ignore
        if log_buffer is None:
            log_buffer = []
        self.__log_buffer = log_buffer

    def getvalue(self):
        if self.__in_mol or self.__in_mol is None:
            raise ValueError('molecule not complete')
        return {'atoms': self.__atoms, 'bonds': self.__bonds, 'stereo': self.__stereo, 'meta': self.__meta,
                'hydrogens': self.__hydrogens}

    def __call__(self, line):
        if self.__in_mol:
            if line.startswith('M  V30 END CTAB'):
                self.__in_mol = False
                return True
            elif self.__atoms_count:
                if line.startswith('M  V30 END'):
                    x = line[11:].strip()
                    cp = self.__parser
                    self.__parser = None

                    if x == 'ATOM':
                        if cp == self.__atom_parser and len(self.__atoms) == self.__atoms_count:
                            return
                    elif x == 'BOND':
                        if cp == self.__bond_parser and len(self.__bonds) == self.__bonds_count:
                            return
                    else:
                        return
                    raise ValueError(f'invalid number of {x} records or invalid CTAB')

                elif self.__parser:
                    collected = self.__record_collector(line)
                    if collected:
                        self.__parser(collected)

                elif line.startswith('M  V30 BEGIN ATOM'):
                    self.__parser = self.__atom_parser
                elif line.startswith('M  V30 BEGIN BOND'):
                    self.__parser = self.__bond_parser
                elif line.startswith('M  V30 BEGIN SGROUP'):
                    self.__parser = self.__sgroup_parser
                elif line.startswith('M  V30 BEGIN COLLECTION'):
                    self.__parser = self.__ignored_block_parser
                else:
                    raise ValueError('invalid CTAB')

            else:  # M  V30 COUNTS line expected
                a, b, *meta = line[13:].split()
                atom_count = int(a)
                if not atom_count:
                    raise EmptyMolecule
                self.__bonds_count = int(b)
                self.__atoms_count = atom_count
                for kv in meta:
                    if '=' in kv:
                        k, v = kv.split('=', 1)
                        if k and v:
                            self.__meta[k] = v

        elif self.__in_mol is not None:
            raise SyntaxError('invalid usage')
        elif not line.startswith('M  V30 BEGIN CTAB'):
            raise ValueError('invalid CTAB')
        else:
            self.__in_mol = True

    def __record_collector(self, line):
        if not line.endswith('-\n'):
            line = line[6:]
            if self.__record:
                line = self.__record + line
                self.__record = None

            line = line.strip()
            collect = []
            tmp = []
            until = None
            for s in line:
                if until:
                    tmp.append(s)
                    if s == until:
                        until = None
                elif s == '(':
                    tmp.append('(')
                    until = ')'
                elif s == '"':
                    tmp.append(s)
                    until = '"'
                elif s == ' ':
                    if tmp:
                        collect.append(''.join(tmp))
                        tmp = []
                else:
                    tmp.append(s)
            if tmp:
                collect.append(''.join(tmp))

            return collect

        line = line[6:-2]
        if not self.__record:
            self.__record = line
        else:
            self.__record += line

    def __bond_parser(self, line):
        _, t, a1, a2, *kvs = line
        if a1 in self.__star_points:
            if a2 in self.__star_points:
                if self.__ignore:
                    self.__log_buffer.append('invalid bond ignored: star-point to star-point')
                    self.__bonds_count -= 1
                    return
                else:
                    raise ValueError('invalid bond: star-point to star-point')
            try:
                star = self.__atom_map[a2]
            except KeyError:
                raise ValueError('invalid atoms number')
            endpoints = None
        elif a2 in self.__star_points:
            try:
                star = self.__atom_map[a1]
            except KeyError:
                raise ValueError('invalid atoms number')
            endpoints = None
        else:
            star = None
            try:
                t = int(t)
                if t in (9, 10):  # added ad-hoc for bond type 9
                    t = 8
                    self.__log_buffer.append('coordinate bond replaced to special')
                self.__bonds.append((self.__atom_map[a1], self.__atom_map[a2], t))
            except KeyError:
                raise ValueError('invalid atoms numbers')

        for kv in kvs:
            k, v = kv.split('=')
            if k == 'CFG':
                if v == '1':
                    self.__stereo.append((self.__atom_map[a1], self.__atom_map[a2], 1))
                elif v == '3':
                    self.__stereo.append((self.__atom_map[a1], self.__atom_map[a2], -1))
                else:
                    self.__log_buffer.append('invalid or unsupported stereo')
            elif k == 'ENDPTS':
                endpoints = v[1:-1].split()
                if len(endpoints) != int(endpoints[0]) + 1:
                    raise ValueError('invalid ENDPTS block')
        if star is not None:
            if endpoints:
                self.__bonds_count += int(endpoints[0]) - 1
                for m in endpoints[1:]:
                    self.__bonds.append((star, self.__atom_map[m], 8))
            elif self.__ignore:
                self.__bonds_count -= 1
                self.__log_buffer.append('Bond ignored. Star atom not allowed as endpoint')
            else:
                raise ValueError('Star atom not allowed')

    def __atom_parser(self, line):
        n, a, x, y, z, m, *kvs = line
        if a.startswith(('[', 'NOT')):
            raise ValueError('list of atoms not supported')
        elif a == '*':
            self.__star_points.append(n)
            self.__atoms_count -= 1
            return

        i = None
        c = 0
        r = False
        for kv in kvs:
            k, v = kv.split('=', 1)
            if k == 'CHG':
                c = int(v)
            elif k == 'MASS':
                i = int(v)
            elif k == 'RAD':
                r = True
        if a == 'D':
            if i:
                raise ValueError('isotope on deuterium atom')
            a = 'H'
            i = 2

        self.__atom_map[n] = len(self.__atoms)
        self.__atoms.append({'element': a, 'isotope': i, 'charge': c, 'is_radical': r,
                             'x': float(x), 'y': float(y), 'z': float(z), 'mapping': int(m)})

    def __sgroup_parser(self, line):
        _, _type, i, *kvs = line
        if _type.startswith('DAT'):
            a = f = d = None
            for kv in kvs:
                k, v = kv.split('=', 1)
                if k == 'ATOMS':
                    a = tuple(self.__atom_map[x] for x in v[1:-1].split()[1:])
                elif k == 'FIELDNAME':
                    f = v.strip('"')
                if k == 'FIELDDATA':
                    d = v.strip('"')
            if a and f and d:
                if f == 'MRV_IMPLICIT_H':
                    self.__hydrogens[a[0]] = int(d[6:])
                else:
                    self.__meta[f'SGROUP DAT {i}'] = (a, f, d)
        elif _type.startswith('SRU'):
            raise ValueError('Polymers not supported')

    def __ignored_block_parser(self, line):
        return

    __record = __atoms_count = __in_mol = __parser = None


__all__ = ['EMOLRead']



# -*- coding: utf-8 -*-
#
#  Copyright 2014-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2014-2019 Timur Madzhidov tmadzhidov@gmail.com features and API discussion
#  Copyright 2014-2019 Alexandre Varnek <varnek@unistra.fr> base idea of CGR approach
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from .containers import *
from .files import *
from .reactor import *
from .utils import *


smiles = SMILESRead.create_parser(ignore=True, remap=False)
xyz = XYZRead.create_parser()


pickle_cache = False  # store cached attributes in pickle


__all__ = ['smiles', 'xyz', 'mdl_mol']

if 'INCHIRead' in locals():
    inchi = INCHIRead.create_parser()
    __all__.append('inchi')



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from lazy_object_proxy import Proxy
from ...periodictable import ListElement


def _rules_single():
    """
    rules without overlapping. these rules can match once to same set of atoms.
    """
    from ...containers import QueryContainer

    raw_rules = []

    #
    # A   H   A     A     H   A
    #  \ / \ /       \  .. \ /
    #   B   B    >>   B     B
    #  / \ / \       / \  .. \
    # A   H   A     A   H     A
    #
    atoms = ({'atom': 'B', 'neighbors': 4, 'hybridization': 1}, {'atom': 'B', 'neighbors': 4, 'hybridization': 1},
             {'atom': 'H', 'neighbors': 2}, {'atom': 'H', 'neighbors': 2})
    bonds = ((1, 3, 1), (1, 4, 1), (2, 3, 1), (2, 4, 1))
    atom_fix = {}
    bonds_fix = ((1, 3, 8), (2, 4, 8))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #      A            A
    #     //           //
    # B - N >> [B-] - [N+]
    #     \            \
    #      A            A
    #
    atoms = ({'atom': 'B'}, {'atom': 'N', 'neighbors': 3, 'hybridization': 2})
    bonds = ((1, 2, 1),)
    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #     A   A             A      A
    #     |   |             |      |
    # A - B = N - A >> A - [B-] - [N+] - A
    #     |   |             |      |
    #     A   A             A      A
    #
    atoms = ({'atom': 'B', 'neighbors': 4, 'hybridization': 2}, {'atom': 'N', 'neighbors': 4, 'hybridization': 2})
    bonds = ((1, 2, 2),)
    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #   A      A       A   A
    #   |      |       |   |
    #  [B-] = [N+] >>  B - N
    #   |      |       |   |
    #   A      A       A   A
    #
    atoms = ({'atom': 'B', 'charge': -1, 'neighbors': 3, 'hybridization': 2},
             {'atom': 'N', 'charge': 1, 'neighbors': 3, 'hybridization': 2})
    bonds = ((1, 2, 2),)
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #        [A-]                 A
    #         |                   |
    # [A-] - [B+3] - [A-] >> A - [B-] - A
    #         |                   |
    #        [A-]                 A
    #
    atoms = ({'atom': 'B', 'charge': 3, 'neighbors': 4}, {'atom': 'A', 'charge': -1}, {'atom': 'A', 'charge': -1},
             {'atom': 'A', 'charge': -1}, {'atom': 'A', 'charge': -1})
    bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 1), (1, 5, 1))
    atom_fix = {1: (-4, None), 2: (1, None), 3: (1, None), 4: (1, None), 5: (1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #        A             A
    #        |             |
    # [A-] - B - A >> A - [B-] - A
    #        |             |
    #        A             A
    #
    atoms = ({'atom': 'B', 'neighbors': 4}, {'atom': 'A', 'charge': -1}, {'atom': 'A'}, {'atom': 'A'}, {'atom': 'A'})
    bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 1), (1, 5, 1))
    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #      A             A
    #      |             |
    # A -  B - A >> A - [B-] - A
    #      |             |
    #      A             A
    #
    atoms = ({'atom': 'B', 'neighbors': 4, 'hybridization': 1},)
    bonds = ()
    atom_fix = {1: (-1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #     A         A
    #     |         |
    #     N    >>  [N+]
    #   / | \     / | \
    #  A  A  A   A  A  A
    #
    atoms = ({'atom': 'N', 'neighbors': 4, 'hybridization': 1},)
    bonds = ()
    atom_fix = {1: (1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #  aci-nitro
    #      O          O
    #     //         //
    # C = N  >> C - [N+]
    #      \         \
    #       OH       [O-]
    #
    atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1}, {'atom': 'C'})
    bonds = ((1, 2, 1), (1, 3, 2), (1, 4, 2))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 4, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #            O               [O-]
    #           //               /
    # [C,N,O] = N  >> [C,N,O] = [N+]
    #           \                \
    #            A                A
    #
    atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1}, {'atom': ListElement(['C', 'N', 'O'])},
             {'atom': 'A'})
    bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # fix aci-nitro
    #       [O-]          [O-]
    #       /             /
    # C = [N+ ] >>  C - [N+]
    #      \             \\
    #       OH            O
    #
    atoms = ({'atom': 'N', 'charge': 1, 'neighbors': 3}, {'atom': 'O', 'charge': -1, 'neighbors': 1},
             {'atom': 'O', 'neighbors': 1}, {'atom': 'C'})
    bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 2))
    atom_fix = {}
    bonds_fix = ((1, 3, 2), (1, 4, 1))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # fix CN(=O)=N(=O)C
    #
    # [N+] = N = O >> [N+] = [N+] - O-
    #        |                |
    #        A                A
    #
    atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'N', 'charge': 1, 'hybridization': 2, 'neighbors': 3}, {'atom': 'A'})
    bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # Fix CN(=O)=N(=N)C
    # [N+] = N = N - ? >> [N+] = [N+] - [N-] - ?
    #        |                    |
    #        A                    A
    #
    atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'N', 'neighbors': (1, 2), 'hybridization': 2},
             {'atom': 'N', 'charge': 1, 'hybridization': 2, 'neighbors': 3}, {'atom': 'A'})
    bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # For N-case not unique!
    #          N             [N-]
    #         //             /
    # [C,N] = N  >> [C,N] = [N+]
    #         \              \
    #          A              A
    #
    atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'N', 'neighbors': (1, 2), 'hybridization': 2},
             {'atom': ListElement(['C', 'N'])}, {'atom': 'A'})
    bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # [N-] - [N+] = O >> N = [N+] - [O-]
    #         |               |
    #         A               A
    #
    atoms = ({'atom': 'N', 'charge': 1, 'neighbors': 3, 'hybridization': 2}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'N', 'charge': -1, 'hybridization': 1, 'neighbors': (1, 2)})
    bonds = ((1, 2, 2), (1, 3, 1))
    atom_fix = {2: (-1, None), 3: (1, None)}
    bonds_fix = ((1, 2, 1), (1, 3, 2))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #         O              [O-]
    #        //              /
    # [A-] - N   >> [A-] - [N+]
    #        \\             \\
    #         O              O
    #
    atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'A', 'charge': -1})
    bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # O : N : O      O = [N+] - [O-]
    #     |      >>       |
    #     A               A
    #
    atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1}, {'atom': 'A'})
    bonds = ((1, 2, 4), (1, 3, 4), (1, 4, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1), (1, 3, 2))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # Nitrite
    #
    #   O        [O-]
    #  //        /
    # [N-]  >>  N
    #  \\       \\
    #   O        O
    #
    atoms = ({'atom': 'N', 'neighbors': 2, 'charge': -1}, {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 2), (1, 3, 2))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # [O,C,N] = N # N >> [O,C,N] = [N+] = [N-]
    #
    atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'neighbors': 1}, {'atom': ListElement(['O', 'C', 'N'])})
    bonds = ((1, 2, 3), (1, 3, 2))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # [C,N,O-] - [N+] # N  >> [C,N,O] = [N+] = [N-]
    #
    atoms = ({'atom': 'N', 'charge': 1, 'neighbors': 2}, {'atom': 'N', 'neighbors': 1},
             {'atom': ListElement(['N', 'C', 'O']), 'charge': -1, 'hybridization': 1})
    bonds = ((1, 2, 3), (1, 3, 1))
    atom_fix = {2: (-1, None), 3: (1, None)}
    bonds_fix = ((1, 2, 2), (1, 3, 2))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #  A - [N+] # N = [N-]  >> A - N = [N+] = [N-]
    #
    atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'charge': 1, 'neighbors': 2},
             {'atom': 'N', 'charge': -1, 'neighbors': 1}, {'atom': 'A'})
    bonds = ((1, 2, 3), (1, 3, 2), (2, 4, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # A - N = N = N >> A - N = [N+] = [N-]
    #
    atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'neighbors': 2, 'hybridization': 2},
             {'atom': 'N', 'neighbors': 1})
    bonds = ((1, 2, 2), (1, 3, 2))
    atom_fix = {1: (1, None), 3: (-1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # A - NH - N # N >> A - N = [N+] = [N-]
    #
    atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'neighbors': 2, 'hybridization': 1},
             {'atom': 'N', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 3))
    atom_fix = {1: (1, None), 3: (-1, None)}
    bonds_fix = ((1, 2, 2), (1, 3, 2))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # [N-] # N = N - A >> [N-] == [N+] == N - A
    #
    atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'neighbors': 1, 'charge': -1},
             {'atom': 'N', 'hybridization': 2, 'neighbors': 2})
    bonds = ((1, 2, 3), (1, 3, 2))
    atom_fix = {1: (1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # [N-] == N # N >> [N-] == [N+] == [N-]
    #
    atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'neighbors': 1, 'charge': -1}, {'atom': 'N', 'neighbors': 1})
    bonds = ((1, 2, 2), (1, 3, 3))
    atom_fix = {1: (1, None), 3: (-1, None)}
    bonds_fix = ((1, 3, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # A - C # N = NH >> A - [CH] = [N+] = [N-]
    #
    atoms = ({'atom': 'C', 'neighbors': (1, 2)}, {'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'neighbors': 1})
    bonds = ((1, 2, 3), (2, 3, 2))
    atom_fix = {2: (1, None), 3: (-1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # note: order dependent
    # A - C # N = [O,N] >> A - C # [N+] - [O,N-]
    #
    atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': ListElement(['O', 'N']), 'hybridization': 2},
             {'atom': 'C', 'neighbors': (1, 2)})
    bonds = ((1, 2, 2), (1, 3, 3))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # [NH2,OH,SH] - N # C - A >> [NH,O,S-] - [N+] # C
    #
    atoms = ({'atom': ListElement(['O', 'N', 'S']), 'neighbors': 1}, {'atom': 'N', 'neighbors': 2},
             {'atom': 'C', 'neighbors': (1, 2)})
    bonds = ((1, 2, 1), (2, 3, 3))
    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # A - [NH] - N # C >> A - [N-] - [N+] # C
    #
    atoms = ({'atom': 'N', 'neighbors': 2, 'hybridization': 1}, {'atom': 'N', 'neighbors': 2},
             {'atom': 'C', 'neighbors': (1, 2)})
    bonds = ((1, 2, 1), (2, 3, 3))
    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # [NH2,OH,SH] - [N+] # [C-] >> [NH,O,S-] - [N+] # [CH]
    #
    atoms = ({'atom': ListElement(['O', 'N', 'S']), 'neighbors': 1}, {'atom': 'N', 'charge': 1, 'neighbors': 2},
             {'atom': 'C', 'charge': -1, 'neighbors': 1})
    bonds = ((1, 2, 1), (2, 3, 3))
    atom_fix = {1: (-1, None), 3: (1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # A - [NH] - [N+] # [C-] >> A - [N-] - [N+] # [CH]
    #
    atoms = ({'atom': 'N', 'neighbors': 2, 'hybridization': 1}, {'atom': 'N', 'charge': 1, 'neighbors': 2},
             {'atom': 'C', 'charge': -1, 'neighbors': 1})
    bonds = ((1, 2, 1), (2, 3, 3))
    atom_fix = {1: (-1, None), 3: (1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # A - N # C >> A - [N+] # [C-]
    #
    atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'C', 'neighbors': 1}, {'atom': 'A'})
    bonds = ((1, 2, 3), (1, 3, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # fix old diazo rule
    #
    # A - [C-] = [N+] = [NH] >> A - [CH] = [N+] = [N-]
    #
    atoms = ({'atom': 'C', 'charge': -1, 'neighbors': (1, 2), 'hybridization': 2},
             {'atom': 'N', 'charge': 1, 'neighbors': 2}, {'atom': 'N', 'neighbors': 1})
    bonds = ((1, 2, 2), (2, 3, 2))
    atom_fix = {1: (1, None), 3: (-1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #    |           |
    #  - N -   >> - [N+] -
    #    \\          |
    #    [O,N]      [O,N-]
    #
    atoms = ({'atom': 'N', 'neighbors': 4, 'hybridization': 2}, {'atom': ListElement(['O', 'N']), 'hybridization': 2})
    bonds = ((1, 2, 2),)
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # N-oxide radical
    #
    #    |         |
    #  - N*  >>  - N
    #    \\        |
    #     O        O*
    #
    atoms = ({'atom': 'N', 'neighbors': 3, 'hybridization': 2, 'is_radical': True}, {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 2),)
    atom_fix = {1: (0, False), 2: (0, True)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # C           C
    #  \           \
    #   N # N >>   [N+] = [N-]
    #  /           /
    # C           C
    #
    atoms = ({'atom': 'N', 'neighbors': 3}, {'atom': 'N', 'neighbors': 1}, {'atom': 'C'}, {'atom': 'C'})
    bonds = ((1, 2, 3), (1, 3, 1), (1, 4, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #  C - N = [N+]  >>  C - [N+] # N
    #
    atoms = ({'atom': 'N', 'neighbors': 2}, {'atom': 'N', 'charge': 1, 'neighbors': 1}, {'atom': 'C'})
    bonds = ((1, 2, 2), (1, 3, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 3),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #  [N+] - [C-] = O  >>  N = C = O
    #
    atoms = ({'atom': 'C', 'charge': -1, 'neighbors': 2},
             {'atom': 'N', 'charge': 1, 'neighbors': (1, 2), 'hybridization': 1}, {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #  N # C - OH  >>  HN = C = O
    #
    atoms = ({'atom': 'C', 'neighbors': 2}, {'atom': 'N', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 3), (1, 3, 1))
    atom_fix = {}
    bonds_fix = ((1, 2, 2), (1, 3, 2))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #  N # C - [O-]  >>  [N-] = C = O
    #
    atoms = ({'atom': 'C', 'neighbors': 2}, {'atom': 'N', 'neighbors': 1}, {'atom': 'O', 'charge': -1, 'neighbors': 1})
    bonds = ((1, 2, 3), (1, 3, 1))
    atom_fix = {2: (-1, None), 3: (1, None)}
    bonds_fix = ((1, 2, 2), (1, 3, 2))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # - [N+] - [O-]  >>  - N = O
    #
    atoms = ({'atom': 'N', 'charge': 1, 'neighbors': 2, 'hybridization': 1},
             {'atom': 'O', 'charge': -1, 'neighbors': 1})
    bonds = ((1, 2, 1),)
    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #  [C+] - N(R2)  >> C = [N+](R2)
    #
    atoms = ({'atom': 'N', 'neighbors': 3, 'hybridization': 1},
             {'atom': 'C', 'charge': 1, 'hybridization': 1, 'neighbors': (1, 2, 3), 'heteroatoms': 1})
    bonds = ((1, 2, 1),)
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #  [C+] = N(R)  >>  C # [N+](R)
    #
    atoms = ({'atom': 'N', 'neighbors': (1, 2), 'hybridization': 2},
             {'atom': 'C', 'charge': 1, 'neighbors': (1, 2), 'hybridization': 2, 'heteroatoms': 1})
    bonds = ((1, 2, 2),)
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 3),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #      [O,S]H    [O,S]
    #      /         //
    # N = C  >> NH - C
    #
    atoms = ({'atom': 'C', 'hybridization': 2}, {'atom': ListElement(['O', 'S']), 'neighbors': 1},
             {'atom': 'N', 'hybridization': 2})
    bonds = ((1, 2, 1), (1, 3, 2))
    atom_fix = {}
    bonds_fix = ((1, 2, 2), (1, 3, 1))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # fix pyridin-2-one. note: only after amide rule
    #
    atoms = ({'atom': 'C', 'neighbors': 3}, {'atom': ListElement(['O', 'S']), 'neighbors': 1},
             {'atom': 'N', 'neighbors': 2}, {'atom': 'A', 'hybridization': 2}, {'atom': 'A', 'hybridization': 2},
             {'atom': 'A', 'hybridization': 2}, {'atom': 'A', 'hybridization': 2})
    bonds = ((1, 2, 2), (1, 3, 1), (1, 4, 1), (3, 7, 1), (4, 5, (1, 2)), (5, 6, (1, 2)), (6, 7, (1, 2)))
    atom_fix = {}
    bonds_fix = ((1, 2, 1), (1, 3, 2))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # todo:
    # [C;a:10][N;H:2][N:3]=[C:4]1[C:5]=,:[C:6][C:7](=[O:1])[C:8]=,:[C:9]1
    # [C;a:10][N;H:2][N:3]=[C:4]1[C:5](=[O:1])[C:6]=,:[C:7]-,:[C:8]=,:[C:9]1

    #
    #       OH          O
    #      /           //
    # C = C    >> C - C
    #      \           \
    #      [O,N]       [O,N]
    #
    atoms = ({'atom': 'O', 'neighbors': 1}, {'atom': ListElement(['O', 'N'])}, {'atom': 'C'}, {'atom': 'C'})
    bonds = ((1, 3, 1), (2, 3, 1), (3, 4, 2))
    atom_fix = {}
    bonds_fix = ((1, 3, 2), (3, 4, 1))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #       A                   A
    #       |                   |
    #  A - [P+] - [O-]  >>  A - P = O
    #       |                   |
    #       A                   A
    #
    atoms = ({'atom': 'P', 'charge': 1, 'neighbors': 4, 'hybridization': 1},
             {'atom': 'O', 'charge': -1, 'neighbors': 1})
    bonds = ((1, 2, 1),)
    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #       A                   A
    #       |                   |
    #  A - [P-] - [C+]  >>  A - P = C
    #       |                   |
    #       A                   A
    #
    atoms = ({'atom': 'P', 'charge': -1, 'neighbors': 4, 'hybridization': 1},
             {'atom': 'C', 'charge': 1, 'neighbors': (1, 2, 3), 'hybridization': 1})
    bonds = ((1, 2, 1),)
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #   F   F         F   F
    #    \ /           \ /
    # F - P - F >> F - [P-] - F
    #    / \           / \
    #   F   F         F   F
    #
    atoms = ({'atom': 'P', 'neighbors': 6}, {'atom': 'F', 'neighbors': 1}, {'atom': 'F', 'neighbors': 1},
             {'atom': 'F', 'neighbors': 1}, {'atom': 'F', 'neighbors': 1}, {'atom': 'F', 'neighbors': 1},
             {'atom': 'F', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 1), (1, 4, 1), (1, 5, 1), (1, 6, 1), (1, 7, 1))
    atom_fix = {1: (-1, None)}
    bonds_fix = ()
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #     O               O
    #     \\              \\
    # A - [P-] - A  >> A - P - A
    #     //              /
    #     O              [O-]
    #
    atoms = ({'atom': 'P', 'charge': -1, 'neighbors': 4}, {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'A'}, {'atom': 'A'})
    bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1), (1, 5, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #           A                   A
    #          /                   /
    # [O-] - [S,Si,Se+]  >>  O = [S,Si,Se]
    #          \                   \
    #           A                   A
    #
    atoms = ({'atom': ListElement(['S', 'Se', 'Si']), 'charge': 1, 'neighbors': 3, 'hybridization': 1},
             {'atom': 'O', 'charge': -1, 'neighbors': 1})
    bonds = ((1, 2, 1),)
    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #       A?                  A?
    #       |                   |
    #  A = [S+] - [O-]  >>  A = S = O
    #       |                   |
    #       A?                  A?
    #
    atoms = ({'atom': 'S', 'charge': 1, 'neighbors': (2, 4), 'hybridization': 2},
             {'atom': 'O', 'charge': -1, 'neighbors': 1})
    bonds = ((1, 2, 1),)
    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #       [O-]          O
    #       /            //
    # A = [S+2]  >>  A = S
    #       \            \\
    #       [O-]          O
    #
    atoms = ({'atom': 'S', 'charge': 2, 'neighbors': 3, 'hybridization': 2},
             {'atom': 'O', 'charge': -1, 'neighbors': 1}, {'atom': 'O', 'charge': -1, 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 1))
    atom_fix = {1: (-2, None), 2: (1, None), 3: (1, None)}
    bonds_fix = ((1, 2, 2), (1, 3, 2))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #          A                  A
    #          |                  |
    # [O-] - [S+2] - [O-] >>  O = S = O
    #          |                  |
    #          A                  A
    #
    atoms = ({'atom': 'S', 'charge': 2, 'neighbors': 4, 'hybridization': 1},
             {'atom': 'O', 'charge': -1, 'neighbors': 1}, {'atom': 'O', 'charge': -1, 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 1))
    atom_fix = {1: (-2, None), 2: (1, None), 3: (1, None)}
    bonds_fix = ((1, 2, 2), (1, 3, 2))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #  A - [S-] - [C+]  >>  A - S = C
    #       |                   |
    #       A                   A
    #
    atoms = ({'atom': 'S', 'charge': -1, 'neighbors': 3, 'hybridization': 1},
             {'atom': 'C', 'charge': 1, 'hybridization': 1, 'neighbors': (1, 2, 3)})
    bonds = ((1, 2, 1),)
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #  O              O
    #  \\             \\
    #  [S-] - A  >>    S - A
    #  //             /
    #  O            [O-]
    #
    atoms = ({'atom': 'S', 'charge': -1, 'neighbors': 3}, {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'A'})
    bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #     O             [O-]
    #     \\             \
    # A - [S-] = O >> A - S = O
    #     //             //
    #     O              O
    #
    atoms = ({'atom': 'S', 'charge': -1, 'neighbors': 4}, {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'O', 'neighbors': 1}, {'atom': 'A'})
    bonds = ((1, 2, 2), (1, 3, 2), (1, 4, 2), (1, 5, 1))
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #  O              [O-]
    #  \\              \
    #  [S-] - [S-] >>   S = S
    #  //              /
    #  O              [O-]
    #
    atoms = ({'atom': 'S', 'charge': -1, 'neighbors': 3}, {'atom': 'S', 'charge': -1, 'neighbors': 1},
             {'atom': 'O', 'neighbors': 1}, {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2), (1, 4, 2))
    atom_fix = {1: (1, None), 2: (1, None), 3: (-1, None), 4: (-1, None)}
    bonds_fix = ((1, 2, 2), (1, 3, 1), (1, 4, 1))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #    A            A
    #    \            \
    # A - S = O >> A - S = O
    #    /            //
    #  [OH]           O
    #
    atoms = ({'atom': 'S', 'neighbors': 4, 'hybridization': 2}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 1), (1, 3, 2))
    atom_fix = {}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #    A            A
    #    \            \
    # A - S = O >> A - S = O
    #    /            //
    #  [NH]           N
    #
    atoms = ({'atom': 'S', 'neighbors': 4, 'hybridization': 2}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'N', 'neighbors': (1, 2), 'hybridization': 1})
    bonds = ((1, 2, 2), (1, 3, 1))
    atom_fix = {}
    bonds_fix = ((1, 3, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #      A{1,3}               A{1,3}
    #      |                    |
    #  N = S - [OH]  >>  [NH] - S = O
    #
    atoms = ({'atom': 'S', 'neighbors': (3, 5), 'hybridization': 2}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'N', 'neighbors': (1, 2), 'hybridization': 2})
    bonds = ((1, 2, 1), (1, 3, 2))
    atom_fix = {}
    bonds_fix = ((1, 2, 2), (1, 3, 1))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # C # C - [O,NH,S]H  >> C=C=[O,NH,S]
    #
    atoms = ({'atom': ListElement(['O', 'S', 'N']), 'neighbors': 1}, {'atom': 'C', 'neighbors': 2},
             {'atom': 'C', 'neighbors': (1, 2)})
    bonds = ((1, 2, 1), (2, 3, 3))
    atom_fix = {}
    bonds_fix = ((1, 2, 2), (2, 3, 2))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    # C # C - [NH]R  >> C=C=NR
    #
    atoms = ({'atom': 'N', 'neighbors': 2, 'hybridization': 1}, {'atom': 'C', 'neighbors': 2},
             {'atom': 'C', 'neighbors': (1, 2)})
    bonds = ((1, 2, 1), (2, 3, 3))
    atom_fix = {}
    bonds_fix = ((1, 2, 2), (2, 3, 2))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # Carbon Monoxide
    #
    # [CX1] = O  >> [С-] # [O+]
    #
    atoms = ({'atom': 'C', 'neighbors': 1, 'is_radical': True}, {'atom': 'O', 'neighbors': 1})
    bonds = ((1, 2, 2),)
    atom_fix = {1: (-1, False), 2: (1, None)}
    bonds_fix = ((1, 2, 3),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # Ozone
    #
    # [O*] -- O -- [O*]  >>  O == [O+] -- [O-]
    #
    atoms = ({'atom': 'O', 'neighbors': 1, 'is_radical': True}, {'atom': 'O', 'neighbors': 2},
             {'atom': 'O', 'neighbors': 1, 'is_radical': True})
    bonds = ((1, 2, 1), (2, 3, 1))
    atom_fix = {1: (0, False), 2: (1, None), 3: (-1, False)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # Allyl radical
    #
    # [C*] -- [C*] -- [C*]  >>  C == C -- [C*]
    #
    atoms = ({'atom': 'C', 'is_radical': True}, {'atom': 'C', 'is_radical': True}, {'atom': 'C', 'is_radical': True})
    bonds = ((1, 2, 1), (2, 3, 1))
    atom_fix = {1: (0, False), 2: (0, False)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    # Br-ion + I-ion
    #
    #        A           A
    #        |           |
    # [Br-].[I+] >> Br - I
    #        |           |
    #        A           A
    #
    atoms = ({'atom': 'Br', 'charge': -1, 'neighbors': 0},
             {'atom': 'I', 'charge': 1, 'neighbors': 2, 'hybridization': 1},)
    bonds = ()
    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 1),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    #
    #  [I+] - [O-] >> I = O
    #
    atoms = ({'atom': 'I', 'charge': 1}, {'atom': 'O', 'charge': -1, 'neighbors': 1})
    bonds = ((1, 2, 1),)
    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ((1, 2, 2),)
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    compiled_rules = []
    for atoms, bonds, atom_fix, bonds_fix in raw_rules:
        q = QueryContainer()
        for a in atoms:
            q.add_atom(**a)
        for n, m, b in bonds:
            q.add_bond(n, m, b)

        any_atoms = [n for n, a in q.atoms() if a.atomic_symbol == 'A' and n not in atom_fix]
        compiled_rules.append((q, atom_fix, bonds_fix, any_atoms))
    return compiled_rules


def _rules_double():
    from ...containers import QueryContainer

    raw_rules = []

    #
    #     [OH]                O
    #      |                 //
    #  N = S = A  >>  [NH] - S = A
    #      |                 |
    #      A                 A
    #
    atoms = ({'atom': 'S', 'neighbors': 4}, {'atom': 'O', 'neighbors': 1},
             {'atom': 'N', 'neighbors': (1, 2), 'hybridization': 2}, {'atom': 'A'}, {'atom': 'A'})
    bonds = ((1, 2, 1), (1, 3, 2), (1, 4, 2), (1, 5, 1))
    atom_fix = {}
    bonds_fix = ((1, 2, 2), (1, 3, 1))
    raw_rules.append((atoms, bonds, atom_fix, bonds_fix))

    compiled_rules = []
    for atoms, bonds, atom_fix, bonds_fix in raw_rules:
        q = QueryContainer()
        for a in atoms:
            q.add_atom(**a)
        for n, m, b in bonds:
            q.add_bond(n, m, b)

        any_atoms = [n for n, a in q.atoms() if a.atomic_symbol == 'A' and n not in atom_fix]
        compiled_rules.append((q, atom_fix, bonds_fix, any_atoms))
    return compiled_rules


single_rules = Proxy(_rules_single)
double_rules = Proxy(_rules_double)


__all__ = ['single_rules', 'double_rules']



# -*- coding: utf-8 -*-
#
#  Copyright 2019-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from functools import cached_property
from itertools import combinations
from logging import info
from typing import Dict, Set, Tuple, Union, TYPE_CHECKING
from .graph import Stereo
from ..morgan import _morgan
from ...exceptions import AtomNotFound, IsChiral, NotChiral


if TYPE_CHECKING:
    from chython import MoleculeContainer


def _pyramid_sign(n, u, v, w):
    #
    #  |   n /
    #  |   |\
    #  |   | \
    #  |  /|  \
    #  | / u---v
    #  |/___\_/___
    #        w
    #
    nx, ny, nz = n
    ux, uy, uz = u
    vx, vy, vz = v
    wx, wy, wz = w

    q1x = ux - nx
    q1y = uy - ny
    q1z = uz - nz
    q2x = vx - nx
    q2y = vy - ny
    q2z = vz - nz
    q3x = wx - nx
    q3y = wy - ny
    q3z = wz - nz

    vol = q1x * (q2y * q3z - q2z * q3y) + q1y * (q2z * q3x - q2x * q3z) + q1z * (q2x * q3y - q2y * q3x)
    if vol > 0:
        return 1
    elif vol < 0:
        return -1
    return 0


def _cis_trans_sign(n, u, v, w):
    # n      w
    #  \    /
    #   u--v
    #  /    \
    # x      x
    nx, ny = n
    ux, uy = u
    vx, vy = v
    wx, wy = w

    q1x = ux - nx
    q1y = uy - ny
    q2x = vx - ux
    q2y = vy - uy
    q3x = wx - vx
    q3y = wy - vy

    # cross vectors
    q1q2z = q1x * q2y - q1y * q2x
    q2q3z = q2x * q3y - q2y * q3x

    dot = q1q2z * q2q3z
    if dot > 0:
        return 1
    elif dot < 0:
        return -1
    return 0


def _allene_sign(n, u, v, w):
    # n    w
    # |   /
    # u--v
    nx, ny, nz = n
    ux, uy = u
    vx, vy = v
    wx, wy, wz = w

    q1x = ux - nx
    q1y = uy - ny
    q1z = -nz
    q2x = vx - ux
    q2y = vy - uy
    q3x = wx - vx
    q3y = wy - vy
    q3z = wz

    # cross vectors
    q1q2x = -q1z * q2y
    q1q2y = q1z * q2x
    q1q2z = q1x * q2y - q1y * q2x
    q2q3x = q2y * q3z
    q2q3y = -q2x * q3z
    q2q3z = q2x * q3y - q2y * q3x

    q1q2q3x = q1q2y * q2q3z - q1q2z * q2q3y
    q1q2q3y = q1q2z * q2q3x - q1q2x * q2q3z

    dot = q1q2q3x * q2x + q1q2q3y * q2y
    if dot > 0:
        return 1
    elif dot < 0:
        return -1
    return 0


class MoleculeStereo(Stereo):
    __slots__ = ()

    def add_wedge(self: 'MoleculeContainer', n: int, m: int, mark: bool, *, clean_cache=True):
        """
        Add stereo data by wedge notation of bonds. Use it for tetrahedrons of allenes.

        :param n: number of atom from which wedge bond started
        :param m: number of atom to which wedge bond coming
        :param mark: up bond is True, down is False
        """
        if n not in self._atoms:
            raise AtomNotFound
        if n in self._atoms_stereo:
            raise IsChiral

        plane = self._plane
        if n in self._chiral_tetrahedrons:
            if m not in self._bonds[n]:
                raise AtomNotFound

            if self._atoms[m].atomic_number == 1:
                s = _pyramid_sign((*plane[m], mark), *((*plane[x], 0) for x in self._stereo_tetrahedrons[n]))
            else:
                order = [(*plane[x], mark if x == m else 0) for x in self._stereo_tetrahedrons[n]]
                if len(order) == 3:
                    s = _pyramid_sign((*plane[n], 0), *order)
                else:
                    s = _pyramid_sign(order[-1], *order[:3])
            if s:
                self._atoms_stereo[n] = s > 0
                if clean_cache:
                    self.flush_cache()
        else:
            c = self._stereo_allenes_centers.get(n)
            if c:
                if c in self._allenes_stereo:
                    raise IsChiral
                elif c not in self._chiral_allenes:
                    raise NotChiral

                order = self._stereo_allenes[c]
                t1, t2 = self._stereo_allenes_terminals[c]
                w = order.index(m)
                if w == 0:
                    m1 = order[1]
                    r = False
                elif w == 1:
                    m1 = order[0]
                    t1, t2 = t2, t1
                    r = False
                elif w == 2:
                    m1 = order[1]
                    r = True
                else:
                    m1 = order[0]
                    t1, t2 = t2, t1
                    r = True
                s = _allene_sign((*plane[m], mark), plane[t1], plane[t2], (*plane[m1], 0))
                if s:
                    self._allenes_stereo[c] = s < 0 if r else s > 0
                    if clean_cache:
                        self.flush_cache()
            else:
                # only tetrahedrons and allenes supported
                raise NotChiral

    def calculate_cis_trans_from_2d(self: 'MoleculeContainer', *, clean_cache=True):
        """
        Calculate cis-trans stereo bonds from given 2d coordinates. Unusable for SMILES and INCHI.
        """
        cis_trans_stereo = self._cis_trans_stereo
        plane = self._plane
        flag = False
        while self._chiral_cis_trans:
            stereo = {}
            for nm in self._chiral_cis_trans:
                n, m = nm
                n1, m1, *_ = self._stereo_cis_trans[nm]
                s = _cis_trans_sign(plane[n1], plane[n], plane[m], plane[m1])
                if s:
                    stereo[nm] = s > 0
            if stereo:
                cis_trans_stereo.update(stereo)
                flag = True
                self.flush_stereo_cache()
            else:
                break
        if flag and clean_cache:
            self.flush_cache()

    def add_atom_stereo(self: 'MoleculeContainer', n: int, env: Tuple[int, ...], mark: bool, *, clean_cache=True):
        """
        Add stereo data for specified neighbors bypass. Use it for tetrahedrons of allenes.

        :param n: number of tetrahedron atom or central atom of allene.
        :param env: numbers of atoms with specified bypass
        :param mark: clockwise or anti bypass.

        See <https://www.daylight.com/dayhtml/doc/theory/theory.smiles.html> and <http://opensmiles.org/opensmiles.html>
        """
        if n not in self._atoms:
            raise AtomNotFound
        if n in self._atoms_stereo or n in self._allenes_stereo:
            raise IsChiral
        if not isinstance(mark, bool):
            raise TypeError('stereo mark should be bool')

        if n in self._chiral_tetrahedrons:
            self._atoms_stereo[n] = self._translate_tetrahedron_sign(n, env, mark)
            if clean_cache:
                self.flush_cache()
        elif n in self._chiral_allenes:
            self._allenes_stereo[n] = self._translate_allene_sign(n, *env, mark)
            if clean_cache:
                self.flush_cache()
        else:  # only tetrahedrons supported
            raise NotChiral

    def add_cis_trans_stereo(self: 'MoleculeContainer', n: int, m: int, n1: int, n2: int, mark: bool, *,
                             clean_cache=True):
        """
        Add stereo data to cis-trans double bonds (not allenes).

        n1/n=m/n2

        :param n: number of starting atom of double bonds chain (alkenes of cumulenes)
        :param m: number of ending atom of double bonds chain (alkenes of cumulenes)
        :param n1: number of neighboring atom of starting atom
        :param n2: number of neighboring atom of ending atom
        :param mark: cis or trans

        See <https://www.daylight.com/dayhtml/doc/theory/theory.smiles.html> and <http://opensmiles.org/opensmiles.html
        """
        atoms = self._atoms
        if n not in atoms or m not in atoms or n1 not in atoms or n2 not in atoms:
            raise AtomNotFound
        if not isinstance(mark, bool):
            raise TypeError('stereo mark should be bool')
        if (n, m) in self._cis_trans_stereo or (m, n) in self._cis_trans_stereo:
            raise IsChiral

        if (n, m) in self._chiral_cis_trans:
            self._cis_trans_stereo[(n, m)] = self._translate_cis_trans_sign(n, m, n1, n2, mark)
            if clean_cache:
                self.flush_cache()
        elif (m, n) in self._chiral_cis_trans:
            self._cis_trans_stereo[(m, n)] = self._translate_cis_trans_sign(m, n, n2, n1, mark)
            if clean_cache:
                self.flush_cache()
        else:
            raise NotChiral

    def flush_stereo_cache(self):
        """
        Flush chiral morgan and chiral centers cache.
        """
        self.__dict__.pop('_chiral_morgan', None)
        self.__dict__.pop('_MoleculeStereo__chiral_centers', None)

    def fix_stereo(self: 'MoleculeContainer'):
        """
        Reset stereo marks.
        """
        if self._atoms_stereo:  # filter tetrahedrons
            stereo_tetrahedrons = self._stereo_tetrahedrons
            atoms_stereo = {k: v for k, v in self._atoms_stereo.items() if k in stereo_tetrahedrons}
            self._atoms_stereo = self_atoms_stereo = {}
        else:
            atoms_stereo = {}

        if self._allenes_stereo:  # filter allenes
            stereo_allenes = self._stereo_allenes
            allenes_stereo = {k: v for k, v in self._allenes_stereo.items() if k in stereo_allenes}
            self._allenes_stereo = self_allenes_stereo = {}
        else:
            allenes_stereo = {}

        if self._cis_trans_stereo:  # filter cis-trans
            stereo_cis_trans = self._stereo_cis_trans
            cis_trans_stereo = {k: v for k, v in self._cis_trans_stereo.items() if k in stereo_cis_trans}
            self._cis_trans_stereo = self_stereo_cis_trans = {}
        else:
            cis_trans_stereo = {}

        old_stereo = len(atoms_stereo) + len(allenes_stereo) + len(cis_trans_stereo)
        while old_stereo:
            chiral_tetrahedrons = self._chiral_tetrahedrons
            chiral_allenes = self._chiral_allenes
            chiral_cis_trans = self._chiral_cis_trans

            tmp = {}
            for n, s in atoms_stereo.items():
                if n in chiral_tetrahedrons:
                    self_atoms_stereo[n] = s
                else:
                    tmp[n] = s
            atoms_stereo = tmp

            tmp = {}
            for n, s in allenes_stereo.items():
                if n in chiral_allenes:
                    self_allenes_stereo[n] = s
                else:
                    tmp[n] = s
            allenes_stereo = tmp

            tmp = {}
            for n, s in cis_trans_stereo.items():
                if n in chiral_cis_trans:
                    self_stereo_cis_trans[n] = s
                else:
                    tmp[n] = s
            cis_trans_stereo = tmp

            fail_stereo = len(atoms_stereo) + len(allenes_stereo) + len(cis_trans_stereo)
            if fail_stereo == old_stereo:
                break
            old_stereo = fail_stereo
            self.flush_stereo_cache()

    @cached_property
    def _wedge_map(self: 'MoleculeContainer'):
        plane = self._plane
        atoms_stereo = self._atoms_stereo
        allenes_centers = self._stereo_allenes_centers
        atoms = self._atoms
        used = set()
        wedge = []
        for n, s in self._allenes_stereo.items():
            env = self._stereo_allenes[n]
            term = self._stereo_allenes_terminals[n]
            order = [(*env[:2], *term), (*env[1::-1], *term[::-1])]
            if env[2]:
                order.append((env[2], env[1], *term))
                order.append((env[1], env[2], *term[::-1]))
            if env[3]:
                order.append((env[3], env[0], *term[::-1]))
                order.append((env[0], env[3], *term))
            order = sorted(order, key=lambda x: (x[0] in atoms_stereo, x[0] in allenes_centers,
                                                 -atoms[x[0]].atomic_number))
            while (order[0][0], order[0][2]) in used:
                order.append(order.pop(0))
            order = order[0]
            used.add((order[2], order[0]))
            s = self._translate_allene_sign(n, *order[:2])
            v = _allene_sign((*plane[order[0]], 1), plane[order[2]], plane[order[3]], (*plane[order[1]], 0))
            if not v:
                info(f'need 2d clean. wedge stereo ambiguous for atom {{{n}}}')
            if s:
                wedge.append((order[2], order[0], v))
            else:
                wedge.append((order[2], order[0], -v))

        for n, s in atoms_stereo.items():
            order = sorted(self._stereo_tetrahedrons[n], key=lambda x: (x in atoms_stereo, x in allenes_centers,
                                                                        -atoms[x].atomic_number, atoms[x].in_ring))
            while (order[0], n) in used:
                order.append(order.pop(0))
            used.add((n, order[0]))

            s = self._translate_tetrahedron_sign(n, order)
            # need recalculation if XY changed
            if len(order) == 3:
                v = _pyramid_sign((*plane[n], 0),
                                  (*plane[order[0]], 1), (*plane[order[1]], 0), (*plane[order[2]], 0))
            else:
                v = _pyramid_sign((*plane[order[3]], 0),
                                  (*plane[order[0]], 1), (*plane[order[1]], 0), (*plane[order[2]], 0))
            if not v:
                info(f'need 2d clean. wedge stereo ambiguous for atom {{{n}}}')
            if s:
                wedge.append((n, order[0], v))
            else:
                wedge.append((n, order[0], -v))
        return tuple(wedge)

    @property
    def _chiral_tetrahedrons(self) -> Set[int]:
        return self.__chiral_centers[0]

    @property
    def _chiral_cis_trans(self) -> Set[Tuple[int, int]]:
        return self.__chiral_centers[1]

    @property
    def _chiral_allenes(self) -> Set[int]:
        return self.__chiral_centers[2]

    @cached_property
    def _chiral_morgan(self: Union['MoleculeContainer', 'MoleculeStereo']) -> Dict[int, int]:
        if not self._atoms_stereo and not self._allenes_stereo and not self._cis_trans_stereo:
            return self.atoms_order
        morgan = self.atoms_order.copy()
        atoms_stereo = set(self._atoms_stereo)
        cis_trans_stereo = set(self._cis_trans_stereo)
        allenes_stereo = set(self._allenes_stereo)
        while True:
            # try iteratively differentiate stereo atoms.
            morgan, atoms_stereo, cis_trans_stereo, allenes_stereo, atoms_groups, cis_trans_groups, allenes_groups = \
                self.__differentiation(morgan, atoms_stereo, cis_trans_stereo, allenes_stereo)
            if not atoms_groups and not cis_trans_groups and not allenes_groups:
                break
            # for some rings differentiation by morgan impossible. try randomly set new weights.
            # sometimes this will lead to pseudo chiral centers and non-unique morgan.
            for group in atoms_groups:
                for n in group[:len(group) // 2]:  # set new weight in half of group randomly.
                    morgan[n] = -morgan[n]
            for group in cis_trans_groups:
                for n, _ in group[:len(group) // 2]:  # set new weight in half of group randomly.
                    morgan[n] = -morgan[n]
            for group in allenes_groups:
                for n in group[:len(group) // 2]:  # set new weight in half of group randomly.
                    morgan[n] = -morgan[n]
            morgan = _morgan(morgan, self.int_adjacency)
        return morgan

    @cached_property
    def _rings_tetrahedrons_linkers(self: 'MoleculeContainer') -> Dict[int, Tuple[int, int, int, int]]:
        """
        Ring-linkers tetrahedrons.

        Values are neighbors in first and second rings.
        """
        out = {}
        tetrahedrons = self._stereo_tetrahedrons
        for n, r in self.atoms_rings.items():
            if n in tetrahedrons:
                for nr, mr in combinations(r, 2):
                    if len(set(nr).intersection(mr)) == 1:
                        ni = nr.index(n)
                        mi = mr.index(n)
                        out[n] = (nr[ni - 1], nr[ni - len(nr) + 1], mr[mi - 1], mr[mi - len(mr) + 1])
                        break
        return out

    @cached_property
    def _rings_tetrahedrons(self: 'MoleculeContainer') -> Dict[int, Union[Tuple[int, int], Tuple[int], Tuple]]:
        """
        Tetrahedrons in rings, except ring-linkers.

        Values are out of ring atoms.
        """
        out = {}
        atoms_rings = self.atoms_rings
        tetrahedrons = self._stereo_tetrahedrons
        points = self._rings_tetrahedrons_linkers
        environment = self.not_special_connectivity
        for n, r in atoms_rings.items():
            if n in tetrahedrons and n not in points:
                out[n] = tuple(environment[n].difference(atoms_rings))
        return out

    @cached_property
    def _rings_cumulenes_linkers(self: 'MoleculeContainer') -> Dict[Tuple[int, int], Tuple[int, int, int, int]]:
        """
        Ring-linkers cumulenes except chords.

        Values are neighbors in first and second rings.
        """
        out = {}
        ar = self.atoms_rings
        chord = self._rings_cumulenes
        for (n, *_, m), (n1, m1, n2, m2) in self._stereo_cumulenes.items():
            if n in ar and m in ar and (n, m) not in chord:
                out[(n, m)] = (n1, n2, m1, m2)
        return out

    @cached_property
    def _rings_cumulenes(self: 'MoleculeContainer') -> Set[Tuple[int, int]]:
        """
        Cumulenes in rings always chiral.
        """
        out = set()
        ar = self.atoms_rings
        for n, *_, m in self._stereo_cumulenes:
            if n in ar and m in ar and not set(ar[n]).isdisjoint(ar[m]):
                out.add((n, m))
        return out

    @cached_property
    def _rings_cumulenes_attached(self: 'MoleculeContainer') -> Dict[Tuple[int, int],
                                                                     Union[Tuple[int, int], Tuple[int]]]:
        """
        Cumulenes attached to rings.

        Values are out of ring atoms.
        """
        ar = self.atoms_rings
        out = {}
        for (n, *_, m), (n1, m1, n2, m2) in self._stereo_cumulenes.items():
            if n in ar:
                if m in ar:
                    continue
                if m2:
                    out[(n, m)] = (m1, m2)
                else:
                    out[(n, m)] = (m1,)
            elif m in ar:
                if n2:
                    out[(n, m)] = (n1, n2)
                else:
                    out[(n, m)] = (n1,)
        return out

    @cached_property
    def __chiral_centers(self: Union['MoleculeStereo', 'MoleculeContainer']):
        atoms_rings = self.atoms_rings
        tetrahedrons = self._stereo_tetrahedrons
        cis_trans = self._stereo_cis_trans
        allenes_centers = self._stereo_allenes_centers
        cis_trans_terminals = self._stereo_cis_trans_terminals
        morgan = self._chiral_morgan

        # find new chiral atoms and bonds.
        # tetrahedron is chiral if all its neighbors are unique.
        chiral_t = {n for n, env in tetrahedrons.items() if len({morgan[x] for x in env}) == len(env)}
        # tetrahedrons-linkers is chiral if in each rings neighbors are unique.
        chiral_t.update(n for n, (n1, n2, m1, m2) in self._rings_tetrahedrons_linkers.items()
                        if morgan[n1] != morgan[n2] and morgan[m1] != morgan[m2])

        # required for axes detection.
        graph = {}
        stereogenic = set()
        pseudo = {}

        # double bond is chiral if neighbors of each terminal atom is unique.
        # ring-linkers and rings-attached also takes into account.
        chiral_c = set()
        chiral_a = set()
        for path, (n1, m1, n2, m2) in self._stereo_cumulenes.items():
            if morgan[n1] != morgan.get(n2, 0) and morgan[m1] != morgan.get(m2, 0):
                n, m = path[0], path[-1]
                if len(path) % 2:
                    chiral_a.add(path[len(path) // 2])
                else:
                    chiral_c.add((n, m))
                stereogenic.add(n)
                stereogenic.add(m)
        # ring cumulenes always chiral. can be already added.
        for nm in self._rings_cumulenes:
            n, m = nm
            if any(len(x) < 8 for x in atoms_rings[n]):  # skip small rings.
                if nm in chiral_c:  # remove already added small rings cumulenes.
                    chiral_c.discard(nm)
                elif (c := allenes_centers[n]) in chiral_a:
                    chiral_a.discard(c)
                continue
            elif nm in cis_trans:
                chiral_c.add(nm)
            else:
                chiral_a.add(allenes_centers[n])
            pseudo[m] = n
            graph[n] = set()
            stereogenic.add(n)

        # find chiral axes. build graph of stereogenic atoms in rings.
        # atoms connected then located in same ring or cumulene.
        for n, env in self._rings_tetrahedrons.items():
            if len(env) == 2:  # one or zero non-ring neighbors stereogenic.
                n1, n2 = env
                if morgan[n1] == morgan[n2]:  # only unique non-ring members required.
                    continue
            graph[n] = set()
            stereogenic.add(n)  # non-linker tetrahedrons in rings - stereogenic.
        for n, (n1, n2, m1, m2) in self._rings_tetrahedrons_linkers.items():
            graph[n] = set()
            if morgan[n1] != morgan[n2] or morgan[m1] != morgan[m2]:
                stereogenic.add(n)  # linkers with at least one unsymmetric ring.
        for n, m in self._rings_cumulenes_linkers:
            graph[n] = {m}
            graph[m] = {n}
            # stereogenic atoms already found.
        for (n, m), env in self._rings_cumulenes_attached.items():
            if len(env) == 2:
                n1, n2 = env
                if morgan[n1] == morgan[n2]:  # only unique non-ring members required.
                    continue
            if n in atoms_rings:
                graph[n] = set()  # non ring endpoints not required.
                stereogenic.add(n)  # mark as stereogenic
            else:
                graph[m] = set()
                stereogenic.add(m)

        if len(graph) > 1:  # add bonds to graph. bonds connects atoms in same rings and terminal atoms of cumulenes.
            for n, ms in graph.items():
                for r in atoms_rings[n]:
                    for m in r:
                        if n == m:
                            continue
                        elif m in graph:
                            ms.add(m)
                        elif m in pseudo and (m := pseudo[m]) != n:
                            ms.add(m)
            # remove not stereogenic terminals.
            while True:
                try:
                    n = next(n for n, ms in graph.items() if not ms or len(ms) == 1 and n not in stereogenic)
                except StopIteration:
                    break
                for m in graph.pop(n):
                    graph[m].discard(n)
            # update chiral atoms.
            for n in graph:
                if n in tetrahedrons:
                    chiral_t.add(n)
                elif n in allenes_centers:
                    chiral_a.add(allenes_centers[n])
                else:
                    chiral_c.add(cis_trans_terminals[n])

        # skip already marked.
        chiral_t.difference_update(self._atoms_stereo)
        chiral_a.difference_update(self._allenes_stereo)
        chiral_c.difference_update(self._cis_trans_stereo)
        return chiral_t, chiral_c, chiral_a

    def __differentiation(self: Union['MoleculeStereo', 'MoleculeContainer'], morgan,
                          atoms_stereo, cis_trans_stereo, allenes_stereo):
        bonds = self.int_adjacency

        tetrahedrons = self._stereo_tetrahedrons
        cis_trans = self._stereo_cis_trans
        allenes = self._stereo_allenes

        translate_tetrahedron = self._translate_tetrahedron_sign
        translate_cis_trans = self._translate_cis_trans_sign
        translate_allene = self._translate_allene_sign

        while True:
            morgan_update = {}
            atoms_groups = []
            cis_trans_groups = []
            allenes_groups = []
            # recalculate morgan weights with taking into account existing stereo marks.
            if atoms_stereo:
                grouped_stereo = defaultdict(list)
                for n in atoms_stereo:
                    grouped_stereo[morgan[n]].append(n)  # collect equal stereo atoms.
                for group in grouped_stereo.values():
                    if not len(group) % 2:  # only even number of equal stereo atoms give new stereo center.
                        # process only truly stereogenic.
                        if len(env := tetrahedrons[group[0]]) == len({morgan[x] for x in env}):
                            s = [n for n in group if translate_tetrahedron(n, sorted(tetrahedrons[n], key=morgan.get))]
                            if 0 < len(s) < len(group):  # RS pair required.
                                for m in s:
                                    morgan_update[m] = -morgan[m]
                            for n in group:  # prevent checks repeating.
                                atoms_stereo.discard(n)
                        else:  # stereo group in rings. unambiguous environment order impossible.
                            atoms_groups.append(group)

            if cis_trans_stereo:
                grouped_stereo = defaultdict(list)
                for nm in cis_trans_stereo:
                    n, m = nm
                    if (mn := morgan[n]) <= (mm := morgan[m]):
                        grouped_stereo[mn].append((n, nm))
                    else:
                        grouped_stereo[mm].append((m, nm))
                for group in grouped_stereo.values():
                    if not len(group) % 2:  # only even number of equal stereo bonds give new stereo center.
                        n1, m1, n2, m2 = cis_trans[group[0][1]]
                        if morgan[n1] != morgan.get(n2, 0) and morgan[m1] != morgan.get(m2, 0):
                            s = []
                            for x, nm in group:
                                n, m = nm
                                n1, m1, n2, m2 = cis_trans[nm]
                                if n2 is None:
                                    a = n1
                                else:
                                    a = min(n1, n2, key=morgan.get)
                                if m2 is None:
                                    b = m1
                                else:
                                    b = min(m1, m2, key=morgan.get)
                                if translate_cis_trans(n, m, a, b):
                                    s.append(x)
                            if 0 < len(s) < len(group):  # RS pair required.
                                for n in s:
                                    morgan_update[n] = -morgan[n]
                                for _, nm in group:
                                    cis_trans_stereo.discard(nm)
                        else:
                            cis_trans_groups.append(group)

            if allenes_stereo:
                grouped_stereo = defaultdict(list)
                for c in allenes_stereo:
                    grouped_stereo[morgan[c]].append(c)
                for group in grouped_stereo.values():
                    if not len(group) % 2:  # only even number of equal stereo bonds give new stereo center.
                        n1, m1, n2, m2 = allenes[group[0]]
                        if morgan[n1] != morgan.get(n2, 0) and morgan[m1] != morgan.get(m2, 0):
                            s = []
                            for c in group:
                                n1, m1, n2, m2 = allenes[c]
                                if n2 is None:
                                    a = n1
                                else:
                                    a = min(n1, n2, key=morgan.get)
                                if m2 is None:
                                    b = m1
                                else:
                                    b = min(m1, m2, key=morgan.get)
                                if translate_allene(c, a, b):
                                    s.append(c)
                            if 0 < len(s) < len(group):  # RS pair required.
                                for c in s:
                                    morgan_update[c] = -morgan[c]
                                for c in group:
                                    allenes_stereo.discard(c)
                        else:
                            allenes_groups.append(group)
            if not morgan_update:
                break
            morgan = _morgan({**morgan, **morgan_update}, bonds)
        return morgan, atoms_stereo, cis_trans_stereo, allenes_stereo, atoms_groups, cis_trans_groups, allenes_groups


__all__ = ['MoleculeStereo']



# -*- coding: utf-8 -*-
#
#  Copyright 2017-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict, namedtuple
from io import StringIO, BytesIO, TextIOWrapper, BufferedIOBase, BufferedReader
from itertools import count
from lxml.etree import iterparse, QName, tostring
from pathlib import Path
from traceback import format_exc
from typing import Union, List, Iterator
from ._mdl import MDLStereo
from ..containers import MoleculeContainer, ReactionContainer
from ..exceptions import EmptyMolecule


parse_error = namedtuple('MRVParseError', ('number', 'json', 'log', 'meta'))
organic_set = {'B', 'C', 'N', 'O', 'P', 'S', 'Se', 'F', 'Cl', 'Br', 'I'}
bond_map = {8: '1" queryType="Any', 4: 'A', 1: '1', 2: '2', 3: '3',
            'Any': 8, 'any': 8, 'A': 4, 'a': 4, '1': 1, '2': 2, '3': 3}


def xml_dict(parent_element, stop_list=None):
    stop_list = set() if stop_list is None else set(stop_list)
    out = {}
    for x, y in parent_element.items():
        y = y.strip()
        if y:
            x = '@%s' % x.strip()
            out[x] = y

    text = []
    if len(parent_element):
        elements_grouped = defaultdict(list)
        for element in parent_element:
            name = QName(element).localname
            if name in stop_list:
                text.append(tostring(element, encoding=str, with_tail=False))
            else:
                elements_grouped[name].append(element)

            if element.tail:
                t = element.tail.strip()
                if t:
                    text.append(t)

        for element_tag, element_group in elements_grouped.items():
            if len(element_group) == 1:
                out[element_tag] = xml_dict(element_group[0], stop_list)
            else:
                out[element_tag] = [xml_dict(x, stop_list) for x in element_group]

    if parent_element.text:
        t = parent_element.text.strip()
        if t:
            text.insert(0, t)
    if text:
        out['$'] = ''.join(text)

    return out


class MRVRead(MDLStereo):
    """
    ChemAxon MRV files reader. works similar to opened file object. support `with` context manager.
    on initialization accept opened in binary mode file, string path to file,
    pathlib.Path object or another binary buffered reader object
    """
    def __init__(self, file, **kwargs):
        """
        :param ignore: Skip some checks of data or try to fix some errors.
        :param remap: Remap atom numbers started from one.
        :param store_log: Store parser log if exists messages to `.meta` by key `ParserLog`.
        :param calc_cis_trans: Calculate cis/trans marks from 2d coordinates.
        :param ignore_stereo: Ignore stereo data.
        """
        if isinstance(file, str):
            self.__file = open(file, 'rb')
            self.__is_buffer = False
        elif isinstance(file, Path):
            self.__file = file.open('rb')
            self.__is_buffer = False
        elif isinstance(file, (BytesIO, BufferedReader, BufferedIOBase)):
            self.__file = file
            self.__is_buffer = True
        else:
            raise TypeError('invalid file. BytesIO, BufferedReader and BufferedIOBase subclasses possible')
        super().__init__(**kwargs)
        self._data = self.__reader()

    def close(self, force=False):
        """
        close opened file

        :param force: force closing of externally opened file or buffer
        """
        if not self.__is_buffer or force:
            self.__file.close()

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    def read(self) -> List[Union[ReactionContainer, MoleculeContainer]]:
        """
        parse whole file

        :return: list of parsed molecules or reactions
        """
        return list(iter(self))

    def __iter__(self) -> Iterator[Union[ReactionContainer, MoleculeContainer]]:
        return (x for x in self._data if not isinstance(x, parse_error))

    def __next__(self) -> Union[ReactionContainer, MoleculeContainer]:
        return next(iter(self))

    def __reader(self) -> Iterator[Union[ReactionContainer, MoleculeContainer, parse_error]]:
        for n, (_, element) in enumerate(iterparse(self.__file, tag='{*}MChemicalStruct')):
            parsed = xml_dict(element)
            element.clear()
            if 'molecule' in parsed and isinstance(parsed['molecule'], dict):
                parsed = parsed['molecule']
                if 'propertyList' in parsed and 'property' in parsed['propertyList']:
                    meta = self.__parse_property(parsed['propertyList']['property'])
                else:
                    meta = {}

                try:
                    record = self.__parse_molecule(parsed)
                except (KeyError, ValueError):
                    self._info(f'record consist errors:\n{format_exc()}')
                    yield parse_error(n, parsed, self._format_log(), meta)
                else:
                    record['meta'].update(meta)
                    try:
                        container = self._convert_molecule(record)
                    except ValueError:
                        self._info(f'record consist errors:\n{format_exc()}')
                        yield parse_error(n, parsed, self._format_log(), meta)
                    else:
                        yield container
            elif 'reaction' in parsed and isinstance(parsed['reaction'], dict):
                parsed = parsed['reaction']
                if 'propertyList' in parsed and 'property' in parsed['propertyList']:
                    meta = self.__parse_property(parsed['propertyList']['property'])
                else:
                    meta = {}

                try:
                    record = self.__parse_reaction(parsed)
                except (KeyError, ValueError):
                    self._info(f'record consist errors:\n{format_exc()}')
                    yield parse_error(n, parsed, self._format_log(), meta)
                else:
                    record['meta'] = meta
                    try:
                        container = self._convert_reaction(record)
                    except ValueError:
                        self._info(f'record consist errors:\n{format_exc()}')
                        yield parse_error(n, parsed, self._format_log(), meta)
                    else:
                        yield container
            else:
                self._info('invalid MDocument')
                yield parse_error(n, parsed, self._format_log(), {})

    def __parse_reaction(self, data):
        reaction = {'reactants': [], 'products': [], 'reagents': []}
        title = data.get('@title')
        if title:
            reaction['title'] = title
        for tag, group in (('reactantList', 'reactants'), ('productList', 'products'), ('agentList', 'reagents')):
            if tag in data and 'molecule' in data[tag]:
                molecule = data[tag]['molecule']
                if isinstance(molecule, dict):
                    molecule = (molecule,)
                for m in molecule:
                    try:
                        reaction[group].append(self.__parse_molecule(m))
                    except EmptyMolecule:
                        if not self._ignore:
                            raise
                        self._info('empty molecule ignored')
        return reaction

    def __parse_property(self, data):
        meta = {}
        if isinstance(data, dict):
            key = data['@title']
            val = data['scalar']['$'].strip()
            if key and val:
                meta[key] = val
            else:
                self._info(f'invalid metadata entry: {data}')
        else:
            for x in data:
                key = x['@title']
                val = x['scalar']['$'].strip()
                if key and val:
                    meta[key] = val
                else:
                    self._info(f'invalid metadata entry: {x}')
        return meta

    def __parse_molecule(self, data):
        atoms, bonds, stereo = [], [], []
        hydrogens = {}
        atom_map = {}
        if 'atom' in data['atomArray']:
            da = data['atomArray']['atom']
            if isinstance(da, dict):
                da = (da,)
            for n, atom in enumerate(da):
                atom_map[atom['@id']] = n
                atoms.append({'element': atom['@elementType'],
                              'isotope': int(atom['@isotope']) if '@isotope' in atom else None,
                              'charge': int(atom.get('@formalCharge', 0)),
                              'is_radical': '@radical' in atom,
                              'mapping': int(atom.get('@mrvMap', 0))})
                if '@z3' in atom:
                    atoms[-1].update(x=float(atom['@x3']), y=float(atom['@y3']), z=float(atom['@z3']))
                else:
                    atoms[-1].update(x=float(atom['@x2']) / 2, y=float(atom['@y2']) / 2, z=0.)
                if '@mrvQueryProps' in atom:
                    raise ValueError('queries unsupported')
                if '@hydrogenCount' in atom:
                    hydrogens[n] = int(atom['@hydrogenCount'])
        else:
            atom = data['atomArray']
            for n, (_id, e) in enumerate(zip(atom['@atomID'].split(), atom['@elementType'].split())):
                atom_map[_id] = n
                atoms.append({'element': e, 'charge': 0, 'mapping': 0, 'isotope': None, 'is_radical': False})
            if '@z3' in atom:
                for a, x, y, z in zip(atoms, atom['@x3'].split(), atom['@y3'].split(), atom['@z3'].split()):
                    a['x'] = float(x)
                    a['y'] = float(y)
                    a['z'] = float(z)
            else:
                for a, x, y in zip(atoms, atom['@x2'].split(), atom['@y2'].split()):
                    a['x'] = float(x) / 2
                    a['y'] = float(y) / 2
                    a['z'] = 0.
            if '@isotope' in atom:
                for a, x in zip(atoms, atom['@isotope'].split()):
                    if x != '0':
                        a['isotope'] = int(x)
            if '@formalCharge' in atom:
                for a, x in zip(atoms, atom['@formalCharge'].split()):
                    if x != '0':
                        a['charge'] = int(x)
            if '@mrvMap' in atom:
                for a, x in zip(atoms, atom['@mrvMap'].split()):
                    if x != '0':
                        a['mapping'] = int(x)
            if '@radical' in atom:
                for a, x in zip(atoms, atom['@radical'].split()):
                    if x != '0':
                        a['is_radical'] = True
            if '@mrvQueryProps' in atom:
                raise ValueError('queries unsupported')
        if not atoms:
            raise EmptyMolecule

        if 'bond' in data['bondArray']:
            db = data['bondArray']['bond']
            if isinstance(db, dict):
                db = (db,)
            for bond in db:
                order = bond_map[bond['@queryType' if '@queryType' in bond else '@order']]
                a1, a2 = bond['@atomRefs2'].split()
                if 'bondStereo' in bond:
                    if '$' in bond['bondStereo']:
                        s = bond['bondStereo']['$']
                        if s == 'H':
                            stereo.append((atom_map[a1], atom_map[a2], -1))
                        elif s == 'W':
                            stereo.append((atom_map[a1], atom_map[a2], 1))
                        else:
                            self._info('invalid or unsupported stereo')
                    else:
                        self._info('incorrect bondStereo tag')
                bonds.append((atom_map[a1], atom_map[a2], order))

        mol = {'atoms': atoms, 'bonds': bonds, 'stereo': stereo, 'meta': {}, 'hydrogens': hydrogens}
        if '@title' in data:
            mol['title'] = data['@title']
        return mol


class MRVWrite:
    """
    ChemAxon MRV files writer. works similar to opened for writing file object. support `with` context manager.
    on initialization accept opened for writing in text mode file, string path to file,
    pathlib.Path object or another buffered writer object
    """
    def __init__(self, file):
        if isinstance(file, str):
            self._file = open(file, 'w')
            self._is_buffer = False
        elif isinstance(file, Path):
            self._file = file.open('w')
            self._is_buffer = False
        elif isinstance(file, (TextIOWrapper, StringIO)):
            self._file = file
            self._is_buffer = True
        else:
            raise TypeError('invalid file. '
                            'TextIOWrapper, StringIO, BytesIO, BufferedReader and BufferedIOBase subclasses possible')
        self.__writable = True

    def close(self, force=False):
        """
        write close tag of MRV file and close opened file

        :param force: force closing of externally opened file or buffer
        """
        if not self.__finalized:
            self._file.write('</cml>\n')
            self.__finalized = True
        if self.__writable:
            self.write = self.__write_closed
            self.__writable = False

        if not self._is_buffer or force:
            self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    @staticmethod
    def __write_closed(_):
        raise ValueError('I/O operation on closed writer')

    def write(self, data: Union[ReactionContainer, MoleculeContainer]):
        """
        write single molecule or reaction into file
        """
        self._file.write('<cml>\n')
        self.__write(data)
        self.write = self.__write

    def __write(self, data):
        if isinstance(data, ReactionContainer):
            buffer = ['<MDocument><MChemicalStruct>']
            if not data._arrow:
                data.fix_positions()

            if data.name:
                buffer.append(f'<reaction title="{data.name}">')
            else:
                buffer.append('<reaction>')

            if data.meta:
                buffer.append('<propertyList>')
                for k, v in data.meta.items():
                    if isinstance(v, str):
                        v = f'<![CDATA[{v}]]>'
                    buffer.append(f'<property title="{k}"><scalar>{v}</scalar></property>')
                buffer.append('</propertyList>')
            c = count(1)
            for i, j in ((data.reactants, 'reactantList'), (data.products, 'productList'),
                         (data.reagents, 'agentList')):
                if not i:
                    continue
                buffer.append(f'<{j}>')
                for n, m in zip(c, i):
                    if m.name:
                        buffer.append(f'<molecule title="{m.name}" molID="m{n}">')
                    else:
                        buffer.append(f'<molecule molID="m{n}">')
                    buffer.append(self.__convert_structure(m))
                    buffer.append('</molecule>')
                buffer.append(f'</{j}>')

            buffer.append(f'<arrow type="DEFAULT" x1="{data._arrow[0] * 2:.4f}" y1="0" '
                          f'x2="{data._arrow[1] * 2:.4f}" y2="0"/>')
            buffer.append('</reaction>')
            self._file.writelines(buffer)
        elif not isinstance(data, MoleculeContainer):
            raise TypeError('MoleculeContainer expected')
        else:
            m = self.__convert_structure(data)
            self._file.write('<MDocument><MChemicalStruct>')

            if data.name:
                self._file.write(f'<molecule title="{data.name}">')
            else:
                self._file.write('<molecule>')

            if data.meta:
                self._file.write('<propertyList>')
                for k, v in data.meta.items():
                    if isinstance(v, str):
                        v = f'<![CDATA[{v}]]>'
                    self._file.write(f'<property title="{k}"><scalar>{v}</scalar></property>')
                self._file.write('</propertyList>')
            self._file.write(m)
            self._file.write('</molecule>')
        self._file.write('</MChemicalStruct></MDocument>\n')

    @staticmethod
    def __convert_structure(g):
        gp = g._plane
        gc = g._charges
        gr = g._radicals
        bg = g._bonds
        hg = g._hydrogens
        hb = g.hybridization

        out = ['<atomArray>']
        for n, atom in g._atoms.items():
            x, y = gp[n]
            ih = hg[n]
            out.append(f'<atom id="a{n}" elementType="{atom.atomic_symbol}" '
                       f'x2="{x * 2:.4f}" y2="{y * 2:.4f}" mrvMap="{n}"')
            if gc[n]:
                out.append(f' formalCharge="{gc[n]}"')
            if gr[n]:
                out.append(' radical="monovalent"')
            if atom.isotope:
                out.append(f' isotope="{atom.isotope}"')
            if ih and (atom.atomic_symbol not in organic_set or hb(n) == 4 and atom.atomic_number in (5, 7, 15)):
                out.append(f' hydrogenCount="{ih}"')
            out.append('/>')
        out.append('</atomArray>')

        out.append('<bondArray>')
        wedge = defaultdict(set)
        for n, (i, j, s) in enumerate(g._wedge_map, start=1):
            out.append(f'<bond id="b{n}" atomRefs2="a{i} a{j}" order="{bond_map[bg[i][j].order]}">'
                       f'<bondStereo>{s == 1 and "W" or "H"}</bondStereo></bond>')
            wedge[i].add(j)
            wedge[j].add(i)
        for n, (i, j, bond) in enumerate(g.bonds(), start=len(out)):
            if j not in wedge[i]:
                out.append(f'<bond id="b{n}" atomRefs2="a{i} a{j}" order="{bond_map[bond.order]}"/>')
        out.append('</bondArray>')
        return ''.join(out)

    __finalized = False


__all__ = ['MRVRead', 'MRVWrite']



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from lazy_object_proxy import Proxy
from ...periodictable import ListElement, AnyMetal


def _rules():
    from ...containers import QueryContainer
    rules = []

    #
    #        R - N - C                 R - N - C
    #           /    ||                   /    ||
    #  A - M = C     || >>    A - M .. [C-]    ||
    #           \    ||                  \\    ||
    #        R - N - C               R - [N+]- C
    #
    q = QueryContainer()
    q.add_atom(AnyMetal())
    q.add_atom('C')
    q.add_atom('N', hybridization=1, neighbors=3, heteroatoms=0)
    q.add_atom('N', hybridization=1, neighbors=3, heteroatoms=0)
    q.add_atom('C', hybridization=2)
    q.add_atom('C', hybridization=2)
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    q.add_bond(2, 4, 1)
    q.add_bond(3, 5, 1)
    q.add_bond(4, 6, 1)
    q.add_bond(5, 6, (1, 2))

    atom_fix = {2: (-1, None), 3: (1, None)}  # atom: (charge diff, new radical state or None)
    bonds_fix = ((1, 2, 8), (2, 3, 2))
    rules.append((q, atom_fix, bonds_fix))

    #
    # cyanide
    #
    q = QueryContainer()
    q.add_atom('C')
    q.add_atom('N')
    q.add_atom(AnyMetal())
    q.add_bond(1, 2, 3)
    q.add_bond(1, 3, 1)

    # NOTE: first fixed atom is metal. required for preventing errors in charge.
    # based on ordered dict nature.
    atom_fix = {3: (1, None), 1: (-1, None)}
    bonds_fix = ((1, 3, 8),)
    rules.append((q, atom_fix, bonds_fix))

    q = QueryContainer()
    q.add_atom('C', charge=-1)
    q.add_atom('N')
    q.add_atom(AnyMetal())
    q.add_bond(1, 2, 3)
    q.add_bond(1, 3, 1)

    atom_fix = {}
    bonds_fix = ((1, 3, 8),)
    rules.append((q, atom_fix, bonds_fix))

    #
    # carbonyl
    #
    q = QueryContainer()
    q.add_atom('C')
    q.add_atom('O')
    q.add_atom(AnyMetal())
    q.add_bond(1, 2, 3)
    q.add_bond(1, 3, 1)

    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ((1, 3, 8),)
    rules.append((q, atom_fix, bonds_fix))

    q = QueryContainer()
    q.add_atom('C', is_radical=True)
    q.add_atom('O')
    q.add_atom(AnyMetal())
    q.add_bond(1, 2, 2)
    q.add_bond(1, 3, 1)

    atom_fix = {1: (-1, False), 2: (1, None)}
    bonds_fix = ((1, 3, 8), (1, 2, 3))
    rules.append((q, atom_fix, bonds_fix))

    q = QueryContainer()
    q.add_atom('C', neighbors=2)
    q.add_atom('O')
    q.add_atom(AnyMetal())
    q.add_bond(1, 2, 2)
    q.add_bond(1, 3, 1)

    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ((1, 3, 8), (1, 2, 3))
    rules.append((q, atom_fix, bonds_fix))

    q = QueryContainer()
    q.add_atom('C')
    q.add_atom('O')
    q.add_atom(AnyMetal())
    q.add_atom(AnyMetal())
    q.add_bond(1, 2, 2)
    q.add_bond(1, 3, 1)
    q.add_bond(1, 4, 1)

    atom_fix = {1: (-1, None), 2: (1, None)}
    bonds_fix = ((1, 3, 8), (1, 4, 8), (1, 2, 3))
    rules.append((q, atom_fix, bonds_fix))

    #
    # Ferrocene covalent uncharged
    #
    q = QueryContainer()
    q.add_atom(AnyMetal())
    q.add_atom('C')
    q.add_atom('C')
    q.add_atom('C')
    q.add_atom('C')
    q.add_atom('C')
    q.add_bond(1, 2, 1); q.add_bond(1, 3, 1); q.add_bond(1, 4, 1); q.add_bond(1, 5, 1); q.add_bond(1, 6, 1)
    q.add_bond(2, 3, 1); q.add_bond(3, 4, 1); q.add_bond(4, 5, 1); q.add_bond(5, 6, 1); q.add_bond(6, 2, 1)

    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 8), (1, 3, 8), (1, 4, 8), (1, 5, 8), (1, 6, 8), (3, 4, 2), (5, 6, 2))
    rules.append((q, atom_fix, bonds_fix))

    #
    # Ferrocene covalent uncharged. invalid valence.
    #
    q = QueryContainer()
    q.add_atom(AnyMetal())
    q.add_atom('C')
    q.add_atom('C')
    q.add_atom('C')
    q.add_atom('C')
    q.add_atom('C')
    q.add_bond(1, 2, 1); q.add_bond(1, 3, 1); q.add_bond(1, 4, 1); q.add_bond(1, 5, 1); q.add_bond(1, 6, 1)
    q.add_bond(2, 3, 1); q.add_bond(3, 4, 2); q.add_bond(4, 5, 1); q.add_bond(5, 6, 2); q.add_bond(6, 2, 1)

    atom_fix = {1: (1, None), 2: (-1, None)}
    bonds_fix = ((1, 2, 8), (1, 3, 8), (1, 4, 8), (1, 5, 8), (1, 6, 8), (3, 4, 2), (5, 6, 2))
    rules.append((q, atom_fix, bonds_fix))

    #
    # Ferrocene covalent radical carbon
    #
    q = QueryContainer()
    q.add_atom(AnyMetal())
    q.add_atom('C', is_radical=True)
    q.add_atom('C', is_radical=True)
    q.add_atom('C', is_radical=True)
    q.add_atom('C', is_radical=True)
    q.add_atom('C', is_radical=True)
    q.add_bond(1, 2, 1); q.add_bond(1, 3, 1); q.add_bond(1, 4, 1); q.add_bond(1, 5, 1); q.add_bond(1, 6, 1)
    q.add_bond(2, 3, 1); q.add_bond(3, 4, 1); q.add_bond(4, 5, 1); q.add_bond(5, 6, 1); q.add_bond(6, 2, 1)

    atom_fix = {1: (1, None), 2: (-1, False), 3: (0, False), 4: (0, False), 5: (0, False), 6: (0, False)}
    bonds_fix = ((1, 2, 8), (1, 3, 8), (1, 4, 8), (1, 5, 8), (1, 6, 8), (3, 4, 2), (5, 6, 2))
    rules.append((q, atom_fix, bonds_fix))

    #
    # Ferrocene covalent charged
    #
    q = QueryContainer()
    q.add_atom(AnyMetal())
    q.add_atom('C', charge=-1)
    q.add_atom('C')
    q.add_atom('C')
    q.add_atom('C')
    q.add_atom('C')
    q.add_bond(1, 2, 1); q.add_bond(1, 3, 1); q.add_bond(1, 4, 1); q.add_bond(1, 5, 1); q.add_bond(1, 6, 1)
    q.add_bond(2, 3, 1); q.add_bond(3, 4, 2); q.add_bond(4, 5, 1); q.add_bond(5, 6, 2); q.add_bond(6, 2, 1)

    atom_fix = {}
    bonds_fix = ((1, 2, 8), (1, 3, 8), (1, 4, 8), (1, 5, 8), (1, 6, 8))
    rules.append((q, atom_fix, bonds_fix))

    #
    # Ferrocene coordinate radical carbon
    #
    q = QueryContainer()
    q.add_atom(AnyMetal())
    q.add_atom('C', is_radical=True)
    q.add_atom('C', is_radical=True)
    q.add_atom('C', is_radical=True)
    q.add_atom('C', is_radical=True)
    q.add_atom('C', is_radical=True)
    q.add_bond(1, 2, 8); q.add_bond(1, 3, 8); q.add_bond(1, 4, 8); q.add_bond(1, 5, 8); q.add_bond(1, 6, 8)
    q.add_bond(2, 3, 1); q.add_bond(3, 4, 1); q.add_bond(4, 5, 1); q.add_bond(5, 6, 1); q.add_bond(6, 2, 1)

    atom_fix = {1: (1, None), 2: (-1, False), 3: (0, False), 4: (0, False), 5: (0, False), 6: (0, False)}
    bonds_fix = ((1, 2, 8), (1, 3, 8), (1, 4, 8), (1, 5, 8), (1, 6, 8), (3, 4, 2), (5, 6, 2))
    rules.append((q, atom_fix, bonds_fix))

    #
    # Allyl complexes
    #
    q = QueryContainer()
    q.add_atom(AnyMetal())
    q.add_atom('C')
    q.add_atom('C')
    q.add_atom('C', is_radical=True)
    q.add_bond(1, 2, 8); q.add_bond(1, 3, 8); q.add_bond(1, 4, 8)
    q.add_bond(2, 3, 2)
    q.add_bond(3, 4, 1)

    atom_fix = {1: (1, None), 4: (-1, False)}
    bonds_fix = ()
    rules.append((q, atom_fix, bonds_fix))

    compiled_rules = []
    for q, atom_fix, bonds_fix in rules:
        any_atoms = [n for n, a in q.atoms() if a.atomic_symbol == 'A' and n not in atom_fix]
        any_atoms.extend(n for n, a in q.atoms() if a.atomic_symbol == 'M')
        compiled_rules.append((q, atom_fix, bonds_fix, any_atoms))
    return compiled_rules


rules = Proxy(_rules)


__all__ = ['rules']



# -*- coding: utf-8 -*-
#
#  Copyright 2019-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Adelia Fatykhova <adelik21979@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import deque
from functools import reduce
from itertools import count, permutations, combinations
from logging import info
from operator import or_
from typing import Iterable, List, Iterator, Tuple, Union
from .base import BaseReactor
from .._functions import lazy_product
from ..containers import QueryContainer, MoleculeContainer, ReactionContainer


class Reactor(BaseReactor):
    """
    Reactor for molecules transformations.
    Generates reaction from input molecules using transformation template.

    Reactor calling transforms reactants to products and
    returns generator of reaction transformations with all
    possible reactions.
    """
    def __init__(self, patterns: Tuple[QueryContainer, ...],
                 products: Tuple[Union[MoleculeContainer, QueryContainer], ...], *,
                 delete_atoms: bool = True, one_shot: bool = True,
                 polymerise_limit: int = 10, automorphism_filter: bool = True):
        """
        :param patterns: search patterns for each reactant
        :param products: resulted structures
        :param delete_atoms: if True atoms exists in reactants but
                            not exists in products will be removed
        :param one_shot: do only single reaction center then True, else do all possible combinations of reactions.
        :param polymerise_limit: limit of self reactions. Make sense than one_shot = False.
        """
        if not patterns or not products:
            raise ValueError('empty template')

        if not all(isinstance(x, QueryContainer) for x in patterns):
            raise TypeError('invalid params')
        elif not all(isinstance(x, (QueryContainer, MoleculeContainer)) for x in products):
            raise TypeError('invalid params')
        self.__patterns = patterns
        products_ = reduce(or_, products)

        self.__one_shot = one_shot
        self.__polymerise_limit = polymerise_limit
        self.__products_atoms = tuple(set(m) for m in products)
        self.__automorphism_filter = automorphism_filter
        super().__init__(reduce(or_, patterns), products_, delete_atoms)

    def __call__(self, structures: Iterable[MoleculeContainer]):
        if any(not isinstance(structure, MoleculeContainer) for structure in structures):
            raise TypeError('only list of Molecules possible')

        len_patterns = len(self.__patterns)
        structures = self.__remap(structures)
        s_nums = set(range(len(structures)))
        if self.__one_shot:
            for chosen in permutations(s_nums, len_patterns):
                ignored = [structures[x] for x in s_nums.difference(chosen)]
                chosen = [structures[x] for x in chosen]
                for new in self.__single_stage(chosen, {x for x in ignored for x in x}):
                    r = ReactionContainer([x.copy() for x in structures], new + [x.copy() for x in ignored])
                    if len(new) > 1:  # try to keep salts
                        r.contract_ions()
                    yield r
        else:
            queue = deque(([structures[x] for x in chosen], [structures[x] for x in s_nums.difference(chosen)], 0)
                          for chosen in permutations(s_nums, len_patterns))
            seen = set()
            while queue:
                chosen, ignored, depth = queue.popleft()
                depth += 1
                for new in self.__single_stage(chosen, {x for x in ignored for x in x}):
                    r = ReactionContainer([x.copy() for x in structures], new + [x.copy() for x in ignored])
                    if len(new) > 1:
                        r.contract_ions()  # try to keep salts
                        if str(r) in seen:
                            continue
                        seen.add(str(r))
                        if len(r.products) != len(ignored) + len(self.__products_atoms):
                            info('ambiguous multicomponent structures. skip multistage processing')
                            yield r
                            continue
                    elif str(r) in seen:
                        continue
                    else:
                        seen.add(str(r))

                    if depth < self.__polymerise_limit:
                        prod = r.products
                        if len_patterns == 1:  # simple case. only products or ignored can be transformed.
                            for i in range(len(prod)):
                                queue.append(([prod[i]], [*prod[:i], *prod[i + 1:]], depth))
                        else:  # one of products molecule combined with previously chosen
                            for chp in combinations(chosen, len_patterns - 1):
                                for i in range(len(prod)):
                                    for ch in permutations(self.__remap((prod[i], *chp)), len_patterns):
                                        queue.append((ch, [*prod[:i], *prod[i + 1:]], depth))
                    yield r

    def __single_stage(self, chosen, ignored) -> Iterator[List[MoleculeContainer]]:
        max_ignored_number = max(ignored, default=0)
        united_chosen = reduce(or_, chosen)
        split = len(self.__products_atoms) > 1
        for match in lazy_product(*(x.get_mapping(y, automorphism_filter=self.__automorphism_filter) for x, y in
                                    zip(self.__patterns, chosen))):
            mapping = match[0]
            for m in match[1:]:
                mapping.update(m)
            new = self._patcher(united_chosen, mapping)
            collision = set(new).intersection(ignored)
            if collision:
                new.remap(dict(zip(collision, count(max(max_ignored_number, max(new)) + 1))))

            if split:
                yield [new.substructure(c) for c in new._connected_components(new._bonds)]
            else:
                yield [new]

    @staticmethod
    def __remap(structures) -> List[MoleculeContainer]:
        checked = []
        checked_atoms = set()
        for structure in structures:
            intersection = set(structure).intersection(checked_atoms)
            if intersection:
                mapping = dict(zip(intersection, count(max(max(checked_atoms), max(structure)) + 1)))
                structure = structure.remap(mapping, copy=True)
                info('some atoms in input structures had the same numbers.\n'
                     f'atoms {list(mapping)} were remapped to {list(mapping.values())}')
            checked_atoms.update(structure)
            checked.append(structure)
        return checked

    def __getstate__(self):
        return {'patterns': self.__patterns, 'products_atoms': self.__products_atoms,
                'polymerise_limit': self.__polymerise_limit, 'one_shot': self.__one_shot,
                'automorphism_filter': self.__automorphism_filter, **super().__getstate__()}

    def __setstate__(self, state):
        if 'split' in state:
            raise ValueError('Reactor pickled with incompatible version.')
        self.__patterns = state['patterns']
        self.__one_shot = state['one_shot']
        self.__polymerise_limit = state['polymerise_limit']
        self.__products_atoms = state['products_atoms']
        self.__automorphism_filter = state['automorphism_filter']
        super().__setstate__(state)


__all__ = ['Reactor']



# -*- coding: utf-8 -*-
#
#  Copyright 2019-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Alexander Nikanshin <17071996sasha@gmail.com>
#  Copyright 2019 Tagir Akhmetshin <tagirshin@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from .element import Element
from .groups import GroupXI
from .periods import PeriodIV, PeriodV, PeriodVI, PeriodVII


class Cu(Element, PeriodIV, GroupXI):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 29

    @property
    def isotopes_distribution(self):
        return {63: 0.6917, 64: 0., 65: 0.3083, 67: 0.}

    @property
    def isotopes_masses(self):
        return {63: 62.929601, 64: 63.929764, 65: 64.927794, 67: 66.927730}

    @property
    def _common_valences(self):
        return 0, 1, 2

    @property
    def _valences_exceptions(self):
        return ((2, False, 0, ()),
                (-1, False, 0, ((1, 'Cl'), (1, 'Cl'))),  # CuCl2^-
                (-3, False, 0, ((1, 'S'), (1, 'S'))))  # CuS2^3- - это характерный комплекс для одновалентной меди

    @property
    def atomic_radius(self):
        return 1.45


class Ag(Element, PeriodV, GroupXI):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 47

    @property
    def isotopes_distribution(self):
        return {107: 0.51839, 109: 0.48161, 110: 0., 111: 0.}

    @property
    def isotopes_masses(self):
        return {107: 106.905093, 109: 108.904756, 110: 109.906107, 111: 110.905291}

    @property
    def _common_valences(self):
        return 0, 1

    @property
    def _valences_exceptions(self):
        return ((-1, False, 0, ((1, 'Cl'), (1, 'Cl'))),  # AgCl2^1-
                (-1, False, 0, ((1, 'O'), (1, 'O'))),  # Ag(OH)2^1-
                (-1, False, 0, ((1, 'S'), (1, 'S'))),  # AgS2^1-
                (-1, False, 0, ((1, 'C'), (1, 'C'))),  # Ag(CN)2^1-
                (1, False, 0, ()))

    @property
    def atomic_radius(self):
        return 1.65


class Au(Element, PeriodVI, GroupXI):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 79

    @property
    def isotopes_distribution(self):
        return {195: 0., 197: 1.0, 198: 0.}

    @property
    def isotopes_masses(self):
        return {195: 194.965035, 197: 196.966552, 198: 197.968244}

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ((1, False, 0, ()), (3, False, 0, ()),
                (0, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (-1, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))))

    @property
    def atomic_radius(self):
        return 1.74


class Rg(Element, PeriodVII, GroupXI):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 111

    @property
    def isotopes_distribution(self):
        return {282: 1.0}

    @property
    def isotopes_masses(self):
        return {282: 282.169127}

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ()

    @property
    def atomic_radius(self):
        return 1.74  # unknown, taken radius of previous element in group


__all__ = ['Cu', 'Ag', 'Au', 'Rg']



# -*- coding: utf-8 -*-
#
#  Copyright 2021 Aleksandr Sizov <murkyrussian@gmail.com>
#  Copyright 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from functools import cached_property
from typing import Dict, TYPE_CHECKING
from ...tautomers._acid import rules as acidic_rules
from ...tautomers._base import rules as basic_rules


if TYPE_CHECKING:
    from ....containers import MoleculeContainer


class Pharmacophore:
    __slots__ = ()
    """
    Matcher of atoms in an molecule to some integer number.
    Each integer represents 10 bit, each bit is enable/disable some pharmacophoric feature:
    | H acceptor | H donor | acidic | basic | positive charged | negative charged | ->
    -> | fluorine | any halogen atom | hydrophobic | aromatic |
    After the constructing, binary number will be converted to 10-based integer
    """

    @cached_property
    def pharmacophores(self: 'MoleculeContainer') -> Dict[int, int]:
        """
        Match id of each atom of some molecule with integer number which define some features using in FCFP
        """
        acid = acidic_rules[:-1]  # discard a rule with halogen acids
        base = basic_rules[:3] + basic_rules[4:]  # there is no need a rule with halogen ions

        atoms = self._atoms
        bonds = self._bonds
        hydrogens = self._hydrogens
        neighbors = self.neighbors
        hybridization = self.hybridization

        acceptors_donors = {n for n, a in atoms.items()
                            if (num := a.atomic_number) in (7, 8) or  # any N,O
                            num == 16 and hybridization(n) == 1 and neighbors(n) <= 2}  # or S-sp3
        halogens = {n for n, atom in atoms.items() if atom.atomic_number in (17, 35, 53) and neighbors(n) < 2}
        charged = {n: c for n, c in self._charges.items() if c}

        bins = (
            acceptors_donors,  # H acceptors
            {n for n in acceptors_donors if hydrogens[n]},  # H donors

            {dct[1] for q in acid for dct in q.get_mapping(self, automorphism_filter=False)},  # acidic
            {dct[1] for q in base for dct in q.get_mapping(self, automorphism_filter=False)},  # basic

            (positive := {n for n, c in charged.items() if c > 0}),  # positive charged
            charged.keys() - positive,  # negative charged

            {n for n, a in atoms.items() if a.atomic_number == 9},  # fluorine
            halogens,  # halogens except fluorine

            {n for n, a in atoms.items() if a.atomic_number == 6 and
             all(atoms[x].atomic_number in (1, 5, 6, 9, 14) for x in bonds[n])},  # hydrophobic

            {n for n in atoms if hybridization(n) == 4}  # aromatic
        )

        out = {idx: 0 for idx in atoms}
        for pos, bin_ in zip((512, 256, 128, 64, 32, 16, 8, 4, 2, 1), bins):
            for idx in bin_:
                out[idx] |= pos
        return out


__all__ = ['Pharmacophore']
